@@ -1,4 +1,4 @@
-/*  $Id: CosmoSmashEngine.cpp,v 1.30 2007/06/20 03:02:43 sarrazip Exp $
+/*  $Id: CosmoSmashEngine.cpp,v 1.42 2008/11/16 05:21:32 sarrazip Exp $
     CosmoSmashEngine.cpp - A space rock shooting video game engine.
 
     cosmosmash - A space rock shooting video game.
@@ -22,6 +22,8 @@
 
 #include "CosmoSmashEngine.h"
 
+#include "Controller.h"
+
 #include <flatzebra/PixmapLoadError.h>
 
 #include <assert.h>
@@ -30,6 +32,7 @@
 #include <string.h>
 #include <ctype.h>
 #include <math.h>
+#include <stdint.h>
 
 #include <map>
 #include <vector>
@@ -74,6 +77,7 @@
 #include "question1.xpm"
 
 using namespace std;
+using namespace flatzebra;
 
 #define __(s) (s)  /* Gettext-like macro */
 
@@ -251,7 +255,8 @@ removeNullElementsFromSpriteList(SpriteList &slist)
 CosmoSmashEngine::CosmoSmashEngine(
 		    long initialScore,
 		    bool _mirrorHyperspace, bool useSound,
-		    bool fullScreen, bool automaticPlay)
+		    bool fullScreen, Controller *_controller,
+		    ostream *_gameScriptStream)
 							throw(int, string)
   : GameEngine(theDrawingPixmapSize, "Cosmosmash", fullScreen),
   		// may throw string exception
@@ -297,9 +302,7 @@ CosmoSmashEngine::CosmoSmashEngine(
     groundHeight(0),
     mountainTopPos(0),
 
-    controller(automaticPlay
-		? static_cast<Controller *>(new AutoController())
-		: static_cast<Controller *>(new UserController())),
+    controller(_controller),
     quitKS(SDLK_ESCAPE),
 
     initScore(initialScore),
@@ -316,8 +319,12 @@ CosmoSmashEngine::CosmoSmashEngine(
 
     starPositions(),
 
-    theSoundMixer(NULL)
+    theSoundMixer(NULL),
+
+    gameScriptStream(_gameScriptStream)
 {
+    assert(controller != NULL);
+
     this->useSound = useSound;
 
     assert(initScore >= 0);
@@ -344,6 +351,12 @@ CosmoSmashEngine::CosmoSmashEngine(
 
 CosmoSmashEngine::~CosmoSmashEngine()
 {
+    // Sprites must be killed to avoid memory leaks.
+    killAllAutomaticCharacters();
+    killSpritesInList(baseBulletSprites);
+
+    delete theSoundMixer;
+    delete controller;
     delete playerSprite;
 }
 
@@ -561,8 +574,7 @@ CosmoSmashEngine::setTimeBeforeNextQuestion()
 void
 CosmoSmashEngine::processKey(SDLKey keysym, bool pressed)
 {
-    if (UserController *uc = dynamic_cast<UserController *>(controller.get()))
-	uc->processKey(keysym, pressed);
+    controller->processKey(keysym, pressed);
 
     quitKS.check(keysym, pressed);
 }
@@ -575,10 +587,15 @@ CosmoSmashEngine::tick()
     if (quitKS.isPressed())
 	return false;
 
+    controller->startOfTick();
+
+    bool ret = true;
+
     if (paused)
     {
-	if (controller->isPauseRequested())
+	if (controller->isResumeRequested())
 	{
+	    registerCommand(COMMAND_RESUME);
 	    paused = false;
 	    displayPauseMessage(false);
 	}
@@ -590,48 +607,54 @@ CosmoSmashEngine::tick()
 	detectCollisions();
 
 	if (!animatePlayer())
-	    return false;
-	
-	animateAutomaticCharacters();
-	restoreBackground();
-	drawSprites();
-
-
-	/*  If the game is in demo mode, the Space bar starts a new game.
-	*/
-	if (numLives == 0)
+	    ret = false;
+	else
 	{
-	    if (controller->isStartRequested(useGameExtensions))
-	    {
-		displayStartMessage(false);
+	    animateAutomaticCharacters();
+	    restoreBackground();
+	    drawSprites();
 
-		killAllAutomaticCharacters();
-		centerPlayerSprite();
-		theScore = initScore;
-		thePeakScore = initScore;
-		numLives = NUM_INIT_LIVES;
 
-		addToScore(0);
-		addToNumLives(0);
-	    }
-	    else 
+	    /*  If the game is in demo mode, the Space bar starts a new game.
+	    */
+	    if (numLives == 0)
 	    {
-		/*  At the end of a game, wait for the last base to have
-		    finished exploding before displaying the start
-		    message again.
-		*/
-		if (explosionSprites.size() == 0)
-		    displayStartMessage(true);
+		if (controller->isStartRequested(useGameExtensions))
+		{
+		    registerCommand(COMMAND_START);
+
+		    displayStartMessage(false);
+
+		    killAllAutomaticCharacters();
+		    centerPlayerSprite();
+		    theScore = initScore;
+		    thePeakScore = initScore;
+		    numLives = NUM_INIT_LIVES;
+
+		    addToScore(0);
+		    addToNumLives(0);
+		}
+		else 
+		{
+		    /*  At the end of a game, wait for the last base to have
+			finished exploding before displaying the start
+			message again.
+		    */
+		    if (explosionSprites.size() == 0)
+			displayStartMessage(true);
+		}
 	    }
+
+	    if (paused)
+		displayPauseMessage(true);
 	}
-
-	if (paused)
-	    displayPauseMessage(true);
     }
 
     controller->endOfTick();
 
-    return true;
+    registerTickEnd();
+
+    return ret;
 }
 
 
@@ -641,7 +664,7 @@ CosmoSmashEngine::fireBullet(Couple pos)
     Couple bbSize = baseBulletPA.getImageSize();
     Sprite *s = new Sprite(baseBulletPA,
 		pos, Couple(0, -PLAYER_BULLET_SPEED), Couple(0, 0),
-		Couple(0, 0), bbSize);
+		Couple(-4, 0), bbSize + Couple(4, 0));
     baseBulletSprites.push_back(s);
 
     if (theLevel >= 5)  // double bullet in higher levels
@@ -681,27 +704,28 @@ CosmoSmashEngine::animatePlayer()
     Returns true if the game must continue, or false to have it stop.
 */
 {
-    if (AutoController *ac = dynamic_cast<AutoController *>(controller.get()))
+    // Inform the controller of the current state of the game:
     {
-	ac->beginGameState();
+	controller->beginGameState();
 
-	ac->setPlayerPos(playerSprite->getCenterPos());
+	controller->setPlayerPos(playerSprite->getCenterPos());
 
 	SpriteList::iterator its;
 	for (its = rockSprites.begin(); its != rockSprites.end(); its++)
-	    ac->addRockPos((*its)->getCenterPos());
+	    controller->addRockPos((*its)->getCenterPos());
 	for (its = spinnerSprites.begin(); its != spinnerSprites.end(); its++)
-	    ac->addSpinnerPos((*its)->getCenterPos());
+	    controller->addSpinnerPos((*its)->getCenterPos());
 	for (its = pulsarSprites.begin(); its != pulsarSprites.end(); its++)
-	    ac->addPulsarPos((*its)->getCenterPos());
+	    controller->addPulsarPos((*its)->getCenterPos());
 	for (its = saucerBulletSprites.begin(); its != saucerBulletSprites.end(); its++)
-	    ac->addPulsarPos((*its)->getCenterPos());
+	    controller->addPulsarPos((*its)->getCenterPos());
 
-	ac->endGameState();
+	controller->endGameState();
     }
 
-    if (controller->isResumeRequested())
+    if (numLives > 0 && controller->isPauseRequested())
     {
+	registerCommand(COMMAND_PAUSE);
 	paused = true;
 	return true;
     }
@@ -712,6 +736,9 @@ CosmoSmashEngine::animatePlayer()
 	return true;
     }
 
+    if (controller->isEndRequested())
+	numLives = 0;
+
     if (numLives == 0)
 	return true;
 
@@ -723,9 +750,11 @@ CosmoSmashEngine::animatePlayer()
     /*  Fire a bullet once in a while:
     */
     int ticksBetweenBullets =
-	   (playerBoostTicks > 0 && playerBoostType == CLOSER_BULLETS ? 3 : 6);
+	   (playerBoostTicks > 0 && playerBoostType == CLOSER_BULLETS ? 2 : 4);
     if (tickCount % ticksBetweenBullets == 0 && controller->isShootingActive())
     {
+	registerCommand(COMMAND_SHOOT);
+
 	Couple bbSize = baseBulletPA.getImageSize();
 	Couple pos = playerPos + Couple((playerSize.x - bbSize.x) / 2, 0);
 
@@ -744,6 +773,8 @@ CosmoSmashEngine::animatePlayer()
 
     if (controller->isHyperspaceRequested())
     {
+	registerCommand(COMMAND_HYPERSPACE);
+
 	Couple size = playerSprite->getSize();
 	int newX;
 	if (mirrorHyperspace)
@@ -767,9 +798,15 @@ CosmoSmashEngine::animatePlayer()
     }
 
     if (controller->isLeftMoveRequested())
+    {
+	registerCommand(COMMAND_MOVE_LEFT);
 	playerSpeed.x = -PLAYER_SPEED;
+    }
     if (controller->isRightMoveRequested())
+    {
+	registerCommand(COMMAND_MOVE_RIGHT);
 	playerSpeed.x = +PLAYER_SPEED;
+    }
 
     if (playerSpeed.isZero())
 	return true;
@@ -926,9 +963,11 @@ CosmoSmashEngine::animateFallingObjects(SpriteList &sprites)
 	    continue;
 	}
 
-	int height = groundPos - s->getLowerRightPos().y;
-		// height is positive when object's bottom is above ground
-	if (height <= - groundHeight)  // if rock low enough, it dies
+	// Get y coordinate of bottom of sprite's collision box:
+	int collBoxBottom = (s->getPos() + s->getCollBoxPos() + s->getCollBoxSize()).y;
+	int height = groundPos - collBoxBottom;
+
+	if (height <= - groundHeight)  // if sprite low enough, it dies
 	{
 	    numLandings++;
 
@@ -1024,7 +1063,7 @@ CosmoSmashEngine::animateAutomaticCharacters()
 	    Couple speed(rnd(-2, +2), rnd(3, 10));
 	    speed = accelerate(speed);
 	    Sprite *s = new Sprite(*pa, pos, speed, Couple(0, 0),
-				    size / 3, size * 2 / 3);
+				    size / 3, size / 3);
 	    spinnerSprites.push_back(s);
 	}
 	else
@@ -1055,7 +1094,7 @@ CosmoSmashEngine::animateAutomaticCharacters()
 	speed = accelerate(speed);
 	Sprite *s = new Sprite(pulsarPA,
 				pos, speed, Couple(0, 0),
-				Couple(2, 2), size - Couple(2, 2));
+				size / 4, size / 2);
 	s->setTimeToLive(4 * 20);
 	pulsarSprites.push_back(s);
     }
@@ -1263,7 +1302,7 @@ CosmoSmashEngine::createExplosionSprite(Couple posOfCenterOfExplosion,
     Couple pos = posOfCenterOfExplosion - size / 2;
     Sprite *s = new Sprite(explosionPA, pos,
 		    Couple(0, 0), Couple(0, 0),
-		    Couple(2, 2), size - Couple(2, 2));
+		    size / 4, size / 2);
     s->setTimeToLive(timeToLive);
     explosionSprites.push_back(s);
     return s;
@@ -1782,7 +1821,7 @@ wu_line(SDL_Surface *surface,
     /* y-major.  Calculate 16-bit fixed point fractional part of a pixel that
        X advances every time Y advances 1 pixel, truncating the result so that
        we won't overrun the endpoint along the X axis */
-    erradj = ((Uint64)dx << 32) / (Uint64)dy;
+    erradj = ((uint64_t)dx << 32) / (uint64_t)dy;
     /* draw all pixels other than the first and last */
     while (--dy) {
       erracctmp = erracc;
@@ -1807,7 +1846,7 @@ wu_line(SDL_Surface *surface,
   /* x-major line.  Calculate 16-bit fixed-point fractional part of a pixel
      that Y advances each time X advances 1 pixel, truncating the result so
      that we won't overrun the endpoint along the X axis. */
-  erradj = ((Uint64)dy << 32) / (Uint64)dx;
+  erradj = ((uint64_t)dy << 32) / (uint64_t)dx;
   /* draw all pixels other than the first and last */
   while (--dx) {
     erracctmp = erracc;
@@ -2084,3 +2123,23 @@ CosmoSmashEngine::playSoundEffect(SoundMixer::Chunk &chunk)
 	}
     }
 }
+
+
+void
+CosmoSmashEngine::registerCommand(Command command)
+{
+    if (gameScriptStream == NULL)
+	return;
+
+    *gameScriptStream << static_cast<int>(command) << ' ';
+}
+
+
+void
+CosmoSmashEngine::registerTickEnd()
+{
+    if (gameScriptStream == NULL)
+	return;
+
+    *gameScriptStream << '\n';
+}
