@@ -1,4 +1,4 @@
-/*  $Id: sdlmain.cpp,v 1.14 2007/06/20 03:02:43 sarrazip Exp $
+/*  $Id: sdlmain.cpp,v 1.23 2008/11/15 03:56:36 sarrazip Exp $
     sdlmain.cpp - main() function for Cosmosmash
 
     cosmosmash - A space rock shooting video game.
@@ -25,10 +25,17 @@
 #endif
 
 #include "CosmoSmashEngine.h"
+#include "UserController.h"
+#include "AutoController.h"
+#include "ScriptController.h"
 
+#include <assert.h>
 #include <stdlib.h>
+#include <string.h>
 #include <errno.h>
 #include <time.h>
+#include <fstream>
+#include <memory>
 
 using namespace std;
 
@@ -50,6 +57,8 @@ static struct option knownOptions[] =
     { "no-sound",          no_argument,       NULL, 'n' },
     { "full-screen",       no_argument,       NULL, 'f' },
     { "auto",              no_argument,       NULL, 'a' },
+    { "record",            required_argument, NULL, 'r' },
+    { "play",              required_argument, NULL, 'p' },
 
     { NULL, 0, NULL, 0 }  // marks the end
 };
@@ -87,6 +96,9 @@ displayHelp()
 "--no-sound           Disable sound effects [default is to enable them]\n"
 "--full-screen        Attempt to use full screen mode [default is window mode]\n"
 "--auto               Automatic mode (the computer plays the game itself)\n"
+"--record=FILENAME    Record the game into the specified file\n"
+"                     (named file is overwritten if it already exists)\n"
+"--play=FILENAME      Replay the game in the named file (written by --record)\n"
 "\n"
     ,
     defaultMillesecondsPerFrame);
@@ -103,7 +115,9 @@ int main(int argc, char *argv[])
     bool mirrorHyperspace = false;
     bool useSound = true;
     bool fullScreen = false;
-    bool automaticPlay = false;
+    enum Mode { USER, AUTO, REPLAY } mode = USER;
+    string recordFilename;  // game recorded only if this is not empty
+    string playFilename;  // only relevant if mode == REPLAY
 
 
     #ifdef HAVE_GETOPT_LONG
@@ -163,7 +177,7 @@ int main(int argc, char *argv[])
 		break;
 
 	    case 'a':
-		automaticPlay = true;
+		mode = AUTO;
 		break;
 
 	    case 'v':
@@ -174,6 +188,15 @@ int main(int argc, char *argv[])
 		displayHelp();
 		return EXIT_SUCCESS;
 
+	    case 'r':
+		recordFilename = optarg;
+		break;
+
+	    case 'p':
+		mode = REPLAY;
+		playFilename = optarg;
+		break;
+
 	    default:
 		displayHelp();
 		return EXIT_FAILURE;
@@ -197,9 +220,93 @@ int main(int argc, char *argv[])
 
     try
     {
+	// Create a controller (i.e., a source of commands)
+	// according to 'mode':
+
+	Controller *controller = NULL;
+	ifstream playFile;  // game script to REPLAY, if applicable
+
+	switch (mode)
+	{
+	    case USER:
+		controller = new UserController();
+		break;
+	    case AUTO:
+		controller = new AutoController();
+		break;
+	    case REPLAY:
+		{
+		    // Open the game script file, checks its format version,
+		    // use its random number seed.
+
+		    playFile.open(playFilename.c_str(), ios::in);
+		    if (!playFile)
+		    {
+			int e = errno;
+			fprintf(stderr, "%s: failed to open file %s: %s\n",
+					    PACKAGE, playFilename.c_str(), strerror(e));
+			return EXIT_FAILURE;
+		    }
+
+		    int requiredFileFormatVersion = 1;
+		    try
+		    {
+			ScriptController *sc = new ScriptController(
+						    playFile,
+						    requiredFileFormatVersion);
+			srand(sc->getSeed());
+			controller = sc;
+		    }
+		    catch (int errorCode)
+		    {
+			switch (errorCode)
+			{
+			    case -1:
+				fprintf(stderr, "%s: failed to read from file %s\n",
+						    PACKAGE, playFilename.c_str());
+				break;
+			    case -2:
+				fprintf(stderr, "%s: invalid data at start of file %s\n",
+						    PACKAGE, playFilename.c_str());
+				break;
+			    default:
+				fprintf(stderr, "%s: invalid file format (requiring %d, got %d)\n",
+						    PACKAGE, requiredFileFormatVersion, errorCode);
+			}
+			return EXIT_FAILURE;
+		    }
+		}
+		break;
+	}
+
+	// Create a file into which the game will be recorded, if requested:
+
+	ofstream gameScriptFile;
+	if (!recordFilename.empty())
+	{
+	    gameScriptFile.open(recordFilename.c_str(), ios::out);
+	    if (!gameScriptFile)
+	    {
+		int e = errno;
+		fprintf(stderr, "%s: failed to create file %s: %s\n",
+				    PACKAGE, recordFilename.c_str(), strerror(e));
+		return EXIT_FAILURE;
+	    }
+	    gameScriptFile << GAME_SCRIPT_FORMAT_VERSION << '\n' << static_cast<int>(seed) << '\n';
+	}
+
+	// Create and run the game engine:
+	// Ownership of the Controller instance is given to the
+	// CosmoSmashEngine instance, which is thus responsible
+	// for destroying it.
+
 	CosmoSmashEngine theCosmoSmashEngine(
 				initScore, mirrorHyperspace, useSound,
-				fullScreen, automaticPlay);
+				fullScreen, controller,
+				recordFilename.empty() ? NULL : &gameScriptFile);
+
+	controller = NULL;  // object now owned by theCosmoSmashEngine
+
 	theCosmoSmashEngine.run(millisecondsPerFrame);
     }
     catch (const string &s)
